@@ -23,6 +23,7 @@ import logging
 import argparse
 from configparser import ConfigParser
 from collections import OrderedDict
+import pyotp
 
 
 logging.basicConfig(level=logging.ERROR, format='%(message)s')
@@ -58,7 +59,8 @@ reserved_fields = {
     'url':'URL',
     'password':'Password',
     'notes':'Notes',
-    'title': 'Title'
+    'title': 'Title',
+    'totp': 'otp'
 }
 
 
@@ -98,7 +100,7 @@ def boolean_input(prompt, default=True):
 def get_group(kp, path):
     if type(path) is str:
         _, path = parse_path(path)
-        log.debug('FIXME: pathlist')
+        log.error('FIXME: pathlist')
     group = kp.find_groups(path=path, first=True)
     if group is None:
         log.error(red("No such group ") + bold('/'.join(path)))
@@ -106,7 +108,7 @@ def get_group(kp, path):
     return group
 def get_entry(kp, path):
     if type(path) is str:
-        log.debug('FIXME: pathlist')
+        log.error('FIXME: pathlist')
         _, path = parse_path(path)
     entry = kp.find_entries(path=path, first=True)
     if entry is None:
@@ -120,10 +122,16 @@ def get_field(entry, field_input):
         sys.exit(1)
     return field
 def no_entry(kp, path):
+    if type(path) is str:
+        log.error('FIXME: pathlist')
+        _, path = parse_path(path)
     if kp.find_entries(path=path, first=True):
         log.error(red("There is already an entry at ") + bold('/'.join(path)))
         sys.exit(1)
 def no_group(kp, path):
+    if type(path) is str:
+        log.error('FIXME: pathlist')
+        _, path = parse_path(path)
     if kp.find_groups(path=path, first=True):
         log.error(red("There is already group at ") + bold('/'.join(path)))
         sys.exit(1)
@@ -463,6 +471,8 @@ def open_database(
 
         # open a specific database in config by name
         elif name is not None:
+            # allow name to start with @
+            name = name.lstrip('@')
             if name not in c.sections():
                 log.error(red("No config section found for " + bold(name)))
                 sys.exit(1)
@@ -521,7 +531,8 @@ def type_entries(args):
 
     Selects an entry using `prog`, then sends the password to the keyboard.
     If `tabbed` is true, both the username and password are typed, separated
-    by a tab"""
+    by a tab.  If `totp` is true, generate and type totp.
+    """
 
     from Xlib.error import DisplayNameError
 
@@ -530,6 +541,7 @@ def type_entries(args):
     except DisplayNameError:
         log.error(red("No X11 session found"))
 
+    # build up a dictionary mapping strings (shown in dmenu) to entries
     entry_texts = {}
 
     # type from all databases
@@ -546,7 +558,7 @@ def type_entries(args):
                         entry_text = '/'.join(entry.path)
                     if args.username:
                         entry_text += " ({})".format(entry.username)
-                    entry_texts[entry_text] = kp
+                    entry_texts[entry_text] = entry
         dmenu_text = '\n'.join(sorted(entry_texts.keys()))
 
     # type from specific database
@@ -557,7 +569,7 @@ def type_entries(args):
                 entry_text = '/'.join(entry.path)
                 if args.username:
                     entry_text += " ({})".format(entry.username)
-                entry_texts[entry_text] = kp
+                entry_texts[entry_text] = entry
         dmenu_text = '\n'.join(sorted(entry_texts.keys()))
 
 
@@ -581,9 +593,11 @@ def type_entries(args):
         log.warning("No path returned by {}".format(args.prog))
         return
 
-    kp = entry_texts[stdout]
-    _, selection_path = parse_path(stdout)
-    selected_entry = get_entry(kp, selection_path)
+    # kp = entry_texts[stdout]
+    # _, selection_path = parse_path(stdout)
+    # selected_entry = get_entry(kp, selection_path)
+
+    selected_entry = entry_texts[stdout]
 
     log.debug("selected_entry:{}".format(selected_entry))
 
@@ -594,26 +608,41 @@ def type_entries(args):
             log.error(bold("xdotool ") + red("not found"))
             sys.exit(1)
 
-    # type out password
+    # type out username/password
     k = Controller()
-    if args.tabbed:
-        if selected_entry.username:
+    # parse OTP field and type
+    if args.totp:
+        otp = None
+        if selected_entry.otp is not None:
+            otp = pyotp.parse_uri(selected_entry.otp)
+        if otp is not None:
             if args.xdotool:
-                call_xdotool(['type', selected_entry.username])
-                call_xdotool(['key', 'Tab'])
+                call_xdotool(['type', otp.now()])
             else:
-                k.type(selected_entry.username)
-                k.press(Key.tab)
-                k.release(Key.tab)
+                k.type(otp.now())
         else:
-            log.warning("Selected entry does not have a username")
-    if selected_entry.password:
-        if args.xdotool:
-            call_xdotool(['type', selected_entry.password])
-        else:
-            k.type(selected_entry.password)
+            log.error(red("Selected entry has no OTP field"))
+            sys.exit(1)
     else:
-        log.warning("Selected entry does not have a password")
+        if args.tabbed:
+            if selected_entry.username:
+                if args.xdotool:
+                    call_xdotool(['type', selected_entry.username])
+                    call_xdotool(['key', 'Tab'])
+                else:
+                    k.type(selected_entry.username)
+                    k.press(Key.tab)
+                    k.release(Key.tab)
+            else:
+                log.warning("Selected entry does not have a username")
+        # type out password only
+        if selected_entry.password:
+            if args.xdotool:
+                call_xdotool(['type', selected_entry.password])
+            else:
+                k.type(selected_entry.password)
+        else:
+            log.warning("Selected entry does not have a password")
 
 
 def show(args):
@@ -621,13 +650,22 @@ def show(args):
 
     kp = open_database(**vars(args))
 
-    entry = get_entry(kp, args.path)
+    _, path = parse_path(args.path)
+    entry = get_entry(kp, path)
+
     # show specified field
     if args.field:
         # handle lowercase field input gracefully
         field = get_field(entry, args.field)
         print(entry._get_string_field(field), end='')
 
+    elif args.totp:
+        if entry.otp is None:
+            log.error(red("Entry has no OTP field"))
+            sys.exit(1)
+        import pyotp
+        print(pyotp.parse_uri(entry.otp).now(), end='')
+
     # otherwise, show all fields
     else:
         print(green("Title: ") + (entry.title or ''))
@@ -638,8 +676,19 @@ def show(args):
             Fore.RESET + Back.RESET
         )
         print(green("URL: ") + (entry.url or ''))
+        if entry.otp is not None:
+            print(green("OTP: ") + (entry.otp or ''))
+            import pyotp
+            import qrcode
+            # generate QR code for seed
+            qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L, border=1)
+            qr.add_data(entry.otp)
+            qr.print_ascii()
+            print(green("OTP Code: ") + pyotp.parse_uri(entry.otp).now())
+        # print custom fields
         for field_name, field_value in entry.custom_properties.items():
             print(green("{}: ".format(field_name)) + str(field_value or ''))
+
         print(green("Created: ") + entry.ctime.isoformat())
         print(green("Modified: ") + entry.mtime.isoformat())
 
@@ -702,6 +751,7 @@ def list_entries(args):
         # if entry, print entry contents
         else:
             args.field = None
+            args.totp = None
             show(args)
 
 
@@ -760,19 +810,16 @@ def add(args):
     _, path = parse_path(args.path)
     [group_path, child_name] = decompose_path(path)
 
-    log.debug("args.path:{}".format(args.path))
-    log.debug("group_path:{} , child_name:{}".format(group_path, child_name))
-
     parent_group = get_group(kp, group_path)
 
     # create a new group
     if args.path.endswith('/'):
-        no_group(kp, args.path)
+        no_group(kp, path)
         kp.add_group(parent_group, child_name)
 
     # create a new entry
     else:
-        no_entry(kp, args.path)
+        no_entry(kp, path)
         username = editable_input('Username')
 
         # use urandom for number generation
@@ -835,10 +882,11 @@ def remove(args):
     """Remove an Entry/Group"""
 
     kp = open_database(**vars(args))
+    _, path = parse_path(args.path)
 
     # remove a group
     if args.path.endswith('/'):
-        group = get_group(kp, args.path)
+        group = get_group(kp, path)
         if len(group.entries) > 0:
             log.error(red("Non-empty group ") + bold(args.path))
             sys.exit(1)
@@ -846,7 +894,7 @@ def remove(args):
 
     # remove an entry
     else:
-        entry = get_entry(kp, args.path)
+        entry = get_entry(kp, path)
         entry.delete()
 
     kp.save()
@@ -856,10 +904,11 @@ def edit(args):
     """Edit fields of an Entry"""
 
     kp = open_database(**vars(args))
+    _, path = parse_path(args.path)
 
     # edit group name
     if args.path.endswith('/'):
-        group = get_group(kp, args.path)
+        group = get_group(kp, path)
 
         if args.set:
             field = args.set[0]
@@ -874,7 +923,7 @@ def edit(args):
             group.name = value
 
     else:
-        entry = get_entry(kp, args.path)
+        entry = get_entry(kp, path)
 
         # edit specific field
         if args.field:
@@ -991,6 +1040,11 @@ def kill(args):
     except (FileNotFoundError, ConnectionRefusedError):
         log.error(red("Background server is not running"))
 
+def restart(args):
+    """Restart background process"""
+
+    kill(args)
+    open_database(**vars(args))
 
 def create_parser():
     """Create argparse object"""
@@ -1035,6 +1089,7 @@ def create_parser():
     show_parser = subparsers.add_parser('show', help="show the contents of an entry")
     show_parser.add_argument('path', metavar='PATH', type=str, help="path to entry")
     show_parser.add_argument('--field', metavar='FIELD', type=str, default=None, help="show the contents of a specific field")
+    show_parser.add_argument('--totp', action='store_true', default=False, help="get entry TOTP")
     show_parser.set_defaults(func=show)
 
     # process args for `edit` command
@@ -1050,6 +1105,7 @@ def create_parser():
     type_parser.add_argument('name', type=str, nargs='?', default=None, help="name of database to type from")
     type_parser.add_argument('--prog', metavar='PROG', default='dmenu', help="dmenu-like program to call for entry selection")
     type_parser.add_argument('--tabbed', action='store_true', default=False, help="type both username and password (tab separated)")
+    type_parser.add_argument('--totp', action='store_true', default=False, help="type entry TOTP")
     type_parser.add_argument('--xdotool', action='store_true', default=False, help="use xdotool for typing passwords")
     type_parser.add_argument('--username', action='store_true', default=False, help="show username in parenthesis during selection")
     type_parser.set_defaults(func=type_entries)
@@ -1084,6 +1140,9 @@ def create_parser():
     kill_parser = subparsers.add_parser('kill', help="kill background process")
     kill_parser.set_defaults(func=kill)
 
+    # process args for `restart` command
+    restart_parser = subparsers.add_parser('restart', help="restart background process")
+    restart_parser.set_defaults(func=restart)
 
     # optional arguments
     parser.add_argument('--debug', action='store_true', default=False, help="enable debug messages")
@@ -1107,7 +1166,7 @@ def main():
         print('Debugging enabled...')
         log.setLevel(logging.DEBUG)
         logging.getLogger('pykeepass_cache').setLevel(logging.DEBUG)
-        logging.getLogger('pykeepass.pykeepass').setLevel(logging.DEBUG)
+        logging.getLogger('pykeepass').setLevel(logging.DEBUG)
 
     try:
         args.func(args)
